import java.util.*;
import java.awt.*;
import javax.swing.*;
import javax.sound.midi.*;

/**
 * Esta clase representa a todas las entidades dentro del juego.
 * Entre otras, incluye a los personajes, items, obstaculos y paredes.
 * @author Ricardo Monascal
 **/
public class MundoDeJuego implements MetaEventListener {

// Constantes de la clase.

	/**
	 * Colores vlidos para las fuentes.
	 **/
	private static final Color [] COLORES = new Color [] {Color.CYAN, Color.MAGENTA, Color.BLUE, Color.GREEN.darker()};
	
	/**
	 * Color del texto para la puntuacin.
	 **/
	private static final Color COLOR_PUNTACION = Color.LIGHT_GRAY;
	
	/**
	 * Cantidad de fuentes independientes que tendr el juego.
	 **/
	private static final int NUM_FUENTES = 3;
	
	/**
	  * Posicion de la fuente.
	  **/
	private static final Vector2D POS_FUENTE = new Vector2D (250,360);
	
	/**
	  * Posicion de la partitura.
	  **/
	private static final Vector2D POS_PUNTUACION = new Vector2D (400,30);
	
	/**
	 * Dificultad del juego que corresponde al nivel FACIL.
	 **/
	private static final int FACIL = 1;
	
	/**
	 * Dificultad del juego que corresponde al nivel MEDIO.
	 **/
	private static final int MEDIO = 2;
	
	/**
	 * Dificultad del juego que corresponde al nivel DIFICIL.
	 **/
	private static final int DIFICIL = 3;
	
	/**
	 * Imagen de fondo del juego.
	 **/
	private static final Image IMAGEN_FONDO = new ImageIcon ("Recursos/Fondo.png").getImage();
	
	/**
	 * Imagen de fondo para el menu.
	 **/
	private static final Image IMAGEN_MENU = new ImageIcon ("Recursos/Fondo_Menu.png").getImage();
	
	/**
	 * Estaoo que representa el menu de juego.
	 **/
	public static final int E_MENU = 0;
	
	/**
	 * Estaoo que representa el juego desarrollandose.
	 **/
	public static final int E_JUEGO = 1;
	
	/**
	 * Estaoo que representa la pause.
	 **/
	public static final int E_PAUSA = 2;

// Campos de la clase.

	/**
	 * Dimensiones del mundo.
	 **/
	private Dimension dimensiones;
	
	/**
	 * Fuentes del juego, sistema de partculas.
	 **/
	private Fuente [] fuentes;
	
	/**
	 * Partitura del juego (guia de colores para el jugador).
	 **/
	private Partitura partitura;
	
	/**
	 * Nivel de dificultad actual del juego.
	 **/
	private int dificultad;
	
	/**
	 * Puntuacin que tiene el jugador.
	 **/
	private int puntuacion;
	
	/**
	 * Reproductor de msica para el juego.
	 **/
	private Reproductor reproductor;
	
	/**
	 * Estado del juego.
	 **/
	private int estado;

// Constructor de la clase.

	/**
	 * Constructor de la clase.
	 * (Por defecto, la dificultad del juego ser FACIL)
	 * @param   dimX   Dimensin horizontal del mundo.
	 * @param   dimY   Dimensin vertical del mundo.
	 **/
	protected MundoDeJuego (int dimX , int dimY) {
		
		dimensiones = new Dimension (dimX, dimY);
		fuentes = new Fuente [NUM_FUENTES];
		organizarFuentes();
		dificultad = DIFICIL;//FACIL;
		puntuacion = 0;
		reproductor = new Reproductor ();
		reproductor.cargarSecuencia("Recursos/Quiero_Ser.mid");
		reproductor.agregarListener(this);
		partitura = new Partitura (reproductor);
		partitura.asignarNumCanales(dificultad);
		estado = E_MENU;
		
	}

// Funciones y metodos de la clase.

	/**
	 * Devuelve el color indicado por el arreglos de colores con el ndice sugerido.
	 * @param   indice   ndice en el arreglo de colores.
	 **/
	public static Color obtenerColor (int indice) {
		
		return COLORES[indice];
		
	}

	/**
	 * Devuelve la altura del mundo.
	 **/
	public int obtenerAltura () {
		
		return dimensiones.height;
		
	}

	/**
	 * Devuelve el ancho del mundo.
	 **/
	public int obtenerAncho () {
		
		return dimensiones.width;
		
	}
	
	/**
	 * Devuelve el estado del juego.
	 **/
	public int obtenerEstado () {
		
		return estado;
		
	}
	
	/**
	 * Inicializa las fuentes en las posiciones, colores y potencias inciales.
	 **/
	public void organizarFuentes () {
		
		int i;
		
		// Prueba
		for (i=0 ; i<NUM_FUENTES ; i++)
			fuentes[i] = new Fuente ();
		
		// Organizacin de la fuente principal
		fuentes[0].agregarChorro(POS_FUENTE, new Vector2D(0.0, -20.0), 2.0);
		fuentes[0].agregarChorro(POS_FUENTE, new Vector2D(1.0, -12.0), 2.0);
		fuentes[0].agregarChorro(POS_FUENTE, new Vector2D(-1.0, -12.0), 2.0);
		fuentes[0].agregarChorro(POS_FUENTE, new Vector2D(2.0, -12.0), 2.0);
		fuentes[0].agregarChorro(POS_FUENTE, new Vector2D(-2.0, -12.0), 2.0);
		fuentes[0].asignarColor(0);
		
		// Organizacin de la fuente secundaria
		fuentes[1].agregarChorro(POS_FUENTE.sumar(Vector2D.IZQUIERDA.mult(130.0)), new Vector2D(6.5, -10.0), 2.0);
		fuentes[1].agregarChorro(POS_FUENTE.sumar(Vector2D.IZQUIERDA.mult(100.0)), new Vector2D(5.0, -10.0), 2.0);
		fuentes[1].agregarChorro(POS_FUENTE.sumar(Vector2D.IZQUIERDA.mult(50.0)), new Vector2D(2.5, -10.0), 2.0);
		fuentes[1].agregarChorro(POS_FUENTE, new Vector2D(0, -10.0), 2.0);
		fuentes[1].agregarChorro(POS_FUENTE.sumar(Vector2D.DERECHA.mult(50.0)), new Vector2D(-2.5, -10.0), 2.0);
		fuentes[1].agregarChorro(POS_FUENTE.sumar(Vector2D.DERECHA.mult(100.0)), new Vector2D(-5.0, -10.0), 2.0);
		fuentes[1].agregarChorro(POS_FUENTE.sumar(Vector2D.DERECHA.mult(130.0)), new Vector2D(-6.5, -10.0), 2.0);
		fuentes[1].asignarColor(1);
		
		// Organizacin de la fuente terciaria.
		fuentes[2].agregarChorro(POS_FUENTE.sumar(Vector2D.IZQUIERDA.mult(150.0).sumar(Vector2D.ABAJO.mult(25))), new Vector2D(0, -12.0), 2.0);
		fuentes[2].agregarChorro(POS_FUENTE.sumar(Vector2D.IZQUIERDA.mult(120.0).sumar(Vector2D.ABAJO.mult(30))), new Vector2D(0, -12.0), 2.0);
		fuentes[2].agregarChorro(POS_FUENTE.sumar(Vector2D.IZQUIERDA.mult(75.0).sumar(Vector2D.ABAJO.mult(35))), new Vector2D(0, -12.0), 2.0);
		fuentes[2].agregarChorro(POS_FUENTE.sumar(Vector2D.IZQUIERDA.mult(25.0).sumar(Vector2D.ABAJO.mult(40))), new Vector2D(0, -12.0), 2.0);
		fuentes[2].agregarChorro(POS_FUENTE.sumar(Vector2D.IZQUIERDA.mult(-25.0).sumar(Vector2D.ABAJO.mult(40))), new Vector2D(0, -12.0), 2.0);
		fuentes[2].agregarChorro(POS_FUENTE.sumar(Vector2D.IZQUIERDA.mult(-75.0).sumar(Vector2D.ABAJO.mult(35))), new Vector2D(0, -12.0), 2.0);
		fuentes[2].agregarChorro(POS_FUENTE.sumar(Vector2D.IZQUIERDA.mult(-120.0).sumar(Vector2D.ABAJO.mult(30))), new Vector2D(0, -12.0), 2.0);
		fuentes[2].agregarChorro(POS_FUENTE.sumar(Vector2D.IZQUIERDA.mult(-150.0).sumar(Vector2D.ABAJO.mult(25))), new Vector2D(0, -12.0), 2.0);
		fuentes[2].asignarColor(2);
		
	}
	
	/**
	 * Asigna el color de una fuente particular
	 * @param   cNum   Indice del color a asignar en los colores disponibles.
	 **/
	public void asignarColorFuente (int cNum) {
		
		int selec;
		
		selec = partitura.obtenerSeleccionado();
		
		switch (dificultad) {
			case FACIL:
				fuentes[0].asignarColor(cNum);
				fuentes[1].asignarColor(cNum);
				fuentes[2].asignarColor(cNum);
				break;
			case MEDIO:
				fuentes[selec].asignarColor(cNum);
				fuentes[2].asignarColor(cNum);
				break;
			case DIFICIL:
				fuentes[selec].asignarColor(cNum);
				break;
		}
		
	}
	
	/**
	 * Modifica el ndice grupo de chorros seleccionado.
	 * @param   mod   Dice si debe aumentar (true) o disminuir (false).
	 **/
	public void modificarSeleccionado (boolean mod) {
		
		partitura.modificarSeleccionado(mod);
		
	}
	
	/**
	 * Asigna una nueva dificultad al juego.
	 * @param   dificultad   Nueva dificultad del juego.
	 **/
	public void asignarDificultad(int dificultad) {
		
		this.dificultad = dificultad;
		
	}
	
	/**
	 * Comienza el juego a partir de la pantalla de menu.
	 **/
	public void comenzarJuego () {
		
		estado = E_JUEGO;
		dificultad = partitura.obtenerSeleccionado()+1;
		reproductor.recargar();
		partitura.asignarNumCanales(dificultad);
		partitura.derivarPartitura(reproductor.obtenerSecuencia());
		
	}
	
	/**
	 * Pausa o despausa el juego.
	 **/
	public void pausar () {
		
		if (estado == E_JUEGO) {
			reproductor.pausar();
			estado = E_PAUSA;
		} else {
			reproductor.reproducir();
			estado = E_JUEGO;
		}
		
	}

	/**
	 * Actualiza este mundo de juego.
	 **/
	public void actualizar () {
		
		int i;
		
		switch (estado) {
			case E_MENU:
			case E_PAUSA:
				break;
			case E_JUEGO:
				for (i=0 ; i<NUM_FUENTES ; i++)
					fuentes[i].actualizar();
				
				partitura.actualizar();
				
				puntuacion += partitura.sumarDiferencias(fuentes);
				break;
		}
		
	}

	/**
	 * Pinta este mundo de juego.
	 * @param   g   Objeto grfico sobre el cual pintar.
	 **/
	public void pintar (Graphics g) {
		
		int i;
		
		g.drawImage(IMAGEN_FONDO, 0, 0, null);
		switch (estado) {
			case E_MENU:
				g.drawImage(IMAGEN_MENU, 0, 0, null);
				
				partitura.pintar(g);
				
				g.setColor(Color.RED);
				g.setFont(g.getFont().deriveFont(24.0f));
				g.drawString("Facil", 120, 490);
				g.drawString("Facil", 300, 490);
				g.drawString("Normal", 120, 530);
				g.drawString("Normal", 300, 530);
				g.drawString("Dificil", 120, 570);
				g.drawString("Dificil", 300, 570);
				break;
			case E_PAUSA:
				g.setColor(Color.RED);
				g.setFont(g.getFont().deriveFont(36.0f));
				g.drawString("PAUSA", 190, 50);
			case E_JUEGO:
				
				for (i=0 ; i<NUM_FUENTES ; i++)
					fuentes[i].pintar(g);
				
				partitura.pintar(g);
				
				g.setFont(g.getFont().deriveFont(12.0f));
				g.setColor(COLOR_PUNTACION);
				g.drawString("Puntuacin:", (int)POS_PUNTUACION.obtenerX(), (int)POS_PUNTUACION.obtenerY());
				g.setFont(g.getFont().deriveFont(24.0f));
				g.drawString(String.format("%06d", puntuacion), (int)POS_PUNTUACION.obtenerX(), (int)POS_PUNTUACION.obtenerY()+20);
				break;
		}
		
	}

// Funciones y mtodos de la interfaz MetaEventListener

	/**
	 * Dice que sucede al suce3der un meta-evento de MIDI.
	 * En esta ocasin se maneja solo para aver cuando termina el audio.
	 * @param   evento   Meta-evento que ha ocurrido.
	 **/
	public void meta (MetaMessage event) {
		
		// Se acaba la cancin.
		if (event.getType() == 47) {
			reproductor.terminar();
			estado = E_MENU;
			dificultad = DIFICIL;
			partitura.limpiar();
			partitura.asignarNumCanales(dificultad);
		}
		
	}

}